#!/usr/bin/env python

"""
    A simple application which calculates
    a couple of thermodynamic quantities
    from a CFour frequency job.
    Written by J. Csontos (csontika@gmail.com)
    Last modified: April, 2011.
"""

import sys
from math import exp,log,pi

def main(args):
    global light,planck,avogadro,boltzman,T298,Po,H2kJ,H2kcal,hbar,P_ver
    light    = 2.99792458E+08 #m/s
    planck   = 6.6260755E-34  #Js
    avogadro = 6.02214199E+23 #particle/mol
    boltzman = 1.380658E-23   #J/K
    T298     = 298.15         #K
    bar      = 100000.0       #Pa
    atm      = 101325.0       #Pa
    H2kJ     = 2625.5001      #kJ/mol 1Hartree/particle to kJ/mol
    H2kcal   = 627.50959      #kcal/mol 1Hartree/particle to kcal/mol
    R        = avogadro*boltzman
    hbar     = planck/(2*pi)

    temp     = float(args[0])    # the first command line argument; temperature in Kelvin
    press    = float(args[1])    # the second command line argument; pressure in Pascal
    P_ver    = 26                # the default python version is assumed to be 2.6 (version 2.4.3 or later is supported)

    if sys.version_info[:3] < (2, 4, 3):
        print ''
        print '!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!'
        print 'Python version 2.4.3 or later is needed.'
        print '!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!'
        print ''
        sys.exit(1)
    else:
        if sys.version_info[:3] < (2, 6, 0):
            P_ver = 24

#########################
# if you uncomment the line below you'll get some contributions due to spin-orbit splitting
# the levels are mostly from the NIST Atomic Spectra Database, http://www.nist.gov/pml/data/asd.cfm

#    spinorbit() 
#########################

#########################

    bottom = False # logical variable; bottom of the harmonic valley

    MW = open_mass('Qmass.txt')
    pg = open_point('Qpoint.txt')
    list_rots = open_rots('Qrots.txt')
    list_freqs, zero_freqs = open_freqs('Qfreqs.txt')

#########################
    linear = False
    if pg[1]=='X':
        linear=True

    if (pg[0]=='C' and (pg[2]=='i' or pg[2]=='s')): # Ci, Cs
        sigma=1
    elif (linear and pg[0]=='D'):                   # DXh
        sigma=2
    elif (linear and pg[0]=='C'):                   # CXh
        sigma=1
    elif (pg[0]=='T'):                              # T, Td, Th
        sigma=12
    elif (pg[0]=='O'):                              # O, Oh
        sigma=24
    elif (pg[0]=='I'):                              # I, Ih
        sigma=60
    elif (pg[0]=='S'):                              # Sn
        sigma=int(pg[1])/2
    elif (pg[0]=='C' and not linear):               # Cn, Cnv, Cnh
        sigma=int(pg[1])
    elif (pg[0]=='D' and not linear):               # Dn, Dnv, Dnh
        sigma=int(pg[1])*2

    PG = pg[0:3]                                    # Point group

    if linear:
        if zero_freqs != 5:
            print ('')
            print ('!!!!!!!!!!!!!!!!!!!!')
            print ('The molecule is linear but the number of')
            print ('translational and rotational degrees of freedom is not 5;')
            print ('it is '+str(zero_freqs)+'.')
            print ('Please, check the frequencies and the geometry.')
            print ('!!!!!!!!!!!!!!!!!!!!')
    else:
        if zero_freqs != 6:
            print ('')
            print ('!!!!!!!!!!!!!!!!!!!!')
            print ('The molecule is not linear but the number of')
            print ('translational and rotational degrees of freedom is not 6;')
            print ('it is '+str(zero_freqs)+'.')
            print ('Please, check the frequencies and the geometry.')
            print ('!!!!!!!!!!!!!!!!!!!!')

#########################
    E_t = E_trans(temp)
    S_t = S_trans(temp,press,MW)
    Cv_t= Cv_trans(temp)

#########################
    E_r  = E_rot(temp,linear)
    Cv_r = Cv_rot(temp,linear)

    rTs=[] # list to contain the rotational constants
    for i in range(len(list_rots)):
        rTs.append(rot_Temp(list_rots[i]))

    S_r=S_rot(q_rot(rTs,temp,sigma,linear),linear)

#########################
    E_v      = 0.0
    E_vib_i  = []
    Cv_v     = 0.0
    Cv_vib_i = []
    S_v      = 0.0
    S_vib_i  = []
    Z_v      = 0.0
    Z_vib_i  = []

    for i in range(len(list_freqs)):
        E_vib_i.append(E_vib(vib_Temp(list_freqs[i]),temp,bottom))
        Cv_vib_i.append(Cv_vib(vib_Temp(list_freqs[i]),temp))
        S_vib_i.append(S_vib(vib_Temp(list_freqs[i]),temp))
        Z_vib_i.append(ZPE(vib_Temp(list_freqs[i])))
        E_v  = E_v+E_vib_i[i]
        Cv_v = Cv_v+Cv_vib_i[i]
        S_v  = S_v+S_vib_i[i]
        Z_v  = Z_v+Z_vib_i[i]
#########################

    print_header(MW,press,temp,PG,sigma)

    if (P_ver == 24):
        printout24('enthalpy','Hartree  ',list_freqs,E_t,E_r,E_v,E_vib_i,bottom,Z_v,Z_vib_i)
        printout24('entropy ','Hartree/K',list_freqs,S_t,S_r,S_v,S_vib_i,bottom,Z_v,Z_vib_i)
        printout24('Heat capacity','Hartree/K',list_freqs,Cv_t,Cv_r,Cv_v,Cv_vib_i,bottom,Z_v,Z_vib_i)
    else:
        printout26('enthalpy','Hartree  ',list_freqs,E_t,E_r,E_v,E_vib_i,bottom,Z_v,Z_vib_i)
        printout26('entropy ','Hartree/K',list_freqs,S_t,S_r,S_v,S_vib_i,bottom,Z_v,Z_vib_i)
        printout26('Heat capacity','Hartree/K',list_freqs,Cv_t,Cv_r,Cv_v,Cv_vib_i,bottom,Z_v,Z_vib_i)

    print_Q(temp,press,MW,rTs,sigma,linear,list_freqs,bottom)
    print_E()

##### end of main ##########



#####  TRANSLATION  #################################
def q_trans(T,p,M):
    return (boltzman*T/p)*(2*pi*0.001*M*boltzman*T/(avogadro*planck**2))**1.5

def E_trans(T):
    return 1.5*(avogadro*boltzman/(1000.0*H2kJ))*T

def Cv_trans(T):
    return 1.5*(avogadro*boltzman/(1000.0*H2kJ))

def S_trans(T,p,M):
    return (avogadro*boltzman/(1000.0*H2kJ))*(log(q_trans(T,p,M))+5.0/2.0)

#####  ROTATION     #################################
def rot_Temp(C_rot):
    return (planck*100.0*light/boltzman)*C_rot #K

def q_rot(rT,T,sigma,LINEAR):
    if LINEAR:
        return (1.0/float(sigma))*(T/rT[0])
    else:
        return ((pi**0.5)/float(sigma))*((T**1.5)/(rT[0]*rT[1]*rT[2])**0.5)

def E_rot(T,LINEAR):
    if LINEAR:
        return 1.0*(avogadro*boltzman/(1000.0*H2kJ))*T
    else:
        return 1.5*(avogadro*boltzman/(1000.0*H2kJ))*T

def Cv_rot(T,LINEAR):
    if LINEAR:
        return 1.0*(avogadro*boltzman/(1000.0*H2kJ))
    else:
        return 1.5*(avogadro*boltzman/(1000.0*H2kJ))

def S_rot(q_r,LINEAR):
    if LINEAR:
        return (avogadro*boltzman/(1000.0*H2kJ))*(log(q_r)+2.0/2.0)
    else:
        return (avogadro*boltzman/(1000.0*H2kJ))*(log(q_r)+3.0/2.0)

#####  VIBRATION   #################################
def vib_Temp(freq):
    return planck*freq*100.0*light/boltzman #K

def q_vib(vT,T,BOTTOM):
    if BOTTOM:
        return exp(-vT/(2.0*T))/(1.0-exp(-vT/T))
    else:
        return 1.0/(1.0-exp(-vT/T))

def E_vib(vT,T,BOTTOM):
    if BOTTOM:
        return (avogadro*boltzman/(1000.0*H2kJ))*vT*(0.5+1.0/(exp(vT/T)-1.0))
    else:
        return (avogadro*boltzman/(1000.0*H2kJ))*vT*(1.0/(exp(vT/T)-1.0))

def Cv_vib(vT,T):
        return (avogadro*boltzman/(1000.0*H2kJ))*(((vT/T)**2)*exp(-vT/T)/((1-exp(-vT/T))**2))

def S_vib(vT,T):
    return (avogadro*boltzman/(1000.0*H2kJ))*(((vT/T)/(exp(vT/T)-1.0))-log(1.0-exp(-vT/T)))

def ZPE(vT):
    return (avogadro*boltzman/(1000.0*H2kJ))*vT*(0.5)

#####  ELECTRONIC   #################################
def q_elec(deg0,deg1,DE1,deg2,DE2,T):
    text=r"q_elec should be the multiplicity\n\
    if the energy of the first excited states is\n\
way above that of the ground state "
    print(text)
    q0=deg0
    q1=deg1*exp(-(planck*100.0*light*DE1)/(boltzman*T))
    q2=deg2*exp(-(planck*100.0*light*DE2)/(boltzman*T))
    q=q0+q1+q2
    q11=(q1/q)*(planck*100.0*light*DE1*avogadro)/(1000.0*2625.5001)
    q12=(q2/q)*(planck*100.0*light*DE2*avogadro)/(1000.0*2625.5001)
    S=((q11+q12)/T)+(avogadro*boltzman/(1000.0*2625.5001))*log(q)
    print('Electronic contrib. to enthalpy: {0:12.8f}'.format(q11+q12))
    print('Electronic partition function:   {0:12.8f}'.format(q))
    print('Electronic contrib. to entropy:  {0:12.8f}'.format(S))

####### Qmass.txt file  #############################
def open_mass(fname):
    file = open(fname)
    mass=0.0

    mass=float(file.readline())

    file.close()
    return mass

####### Qpoint.txt file #############################
def open_point(fname):
    file = open(fname)

    pg=file.readline()
    pg=pg[0:3]

    file.close()
    return pg

####### Qrots.txt file ##############################
def open_rots(fname):
    file = open(fname)
    size = int(file.readline())
    rots = []

    for i in range(size):
        rots.append(float(file.readline()))

    file.close()
    return rots

####### Qfreqs.txt file #############################
def open_freqs(fname):

    check = False
    file = open(fname)
    size = int(file.readline())
    freqs = []
    zero_freqs = 0

    for i in range(size):
        line = file.readline()
        if (line.find('i') > 0) :
            line = line.replace('i','0')
            if float(line) > 10.0 :
                check = True
                line = '0.0'

        if float(line) > 10.0 :
            freqs.append(float(line))
        else:
            zero_freqs = zero_freqs + 1

    file.close()
    if check :
        print ('')
        print ('!!!!!!!!!!!!!!!!!!!!')
        print ('There was at least one imaginary frequency with Im(freq) larger than 10.0,')
        print ('and although it was not included in the analysis below ')
        print ('the results probably are meaningless.')
        print ('Please, check the frequencies and the geometry.')
        print ('!!!!!!!!!!!!!!!!!!!!')

    return freqs, zero_freqs
###################################################

def print_header(MW,press,temp,PG,sigma):
    print('')

    if (P_ver == 24):
        print('%25s %.4f %s' % ('Molecular weight: ',MW,'g/mol'))
        print('%25s %.0f %s' % ('Pressure: ',press,'Pa'))
        print('%25s %.2f %s' % ('Temperature: ',temp,'K'))
        print('%25s %s'      % ('Point group: ',PG))
        print('%25s %s'      % ('Sigma: ',sigma))
    else:
        print('{0:25} {1:.4f} {2}'.format('Molecular weight: ',MW,'g/mol'))
        print('{0:25} {1:.0f} {2}'.format('Pressure: ',press,'Pa'))
        print('{0:25} {1:.2f} {2}'.format('Temperature: ',temp,'K'))
        print('{0:25} {1}'.format('Point group: ',PG))
        print('{0:25} {1}'.format('Sigma: ',sigma))

    print('')
### end of print_header ###

def printout24(name,dim,freqs,trans,rot,vib,f,BOTTOM,zpe,z):

    if( name == 'enthalpy' ):
        factor    = 1.0
        dimension = 'kJ/mol'
        print('  @@@ Thermal corrections to %s %s' % (name,'@@@'))
    elif( name == 'entropy ' ):
        factor    = 1000.0
        dimension = 'J/K.mol'
        print('  @@@ Thermal corrections to %s %s' % (name,'@@@'))
    elif( name == 'Heat capacity' ):
        factor    = 1000.0
        dimension = 'J/K.mol'
        print('  @@@ %s %s' % (name,'(constant volume) @@@'))

    print('%s %12.8f %s %9.3f %s' % (' -- Translational: ',trans,dim,factor*H2kJ*trans,dimension))
    print('%s %12.8f %s %9.3f %s' % (' -- Rotational:    ',rot  ,dim,factor*H2kJ*rot  ,dimension))
    print(              '%s' % (' -- Vibrational:   '))
    if BOTTOM:
        print('      from the bottom of the valley')
    else:
        print('      from the first frequency')
    print('  *** individual frequency contrib. ***')
    print('')


    for i in range(len(freqs)):
        print('%12.2f %s %14.8f %s %9.3f %s' % (freqs[i],'1/cm',f[i],dim,factor*H2kJ*f[i],dimension))
    print('')
    print('%s %12.8f %s %9.3f %s' % ('  *** total vib:   ',vib,dim,factor*H2kJ*vib,dimension))
    print('')

    print('%s %12.8f %s %9.3f %s' % ('  @@@ Total:       ',trans+rot+vib,dim,factor*H2kJ*(trans+rot+vib),dimension))
    print('  ----------------------------------------------------------')
    if( name == 'enthalpy'):
        print(              '%s' % (' -- ZPE:           '))
        print('  *** individual frequency contrib. ***')
        print('')

        for i in range(len(freqs)):
            print('%12.2f %s %14.8f %s %9.3f %s' % (freqs[i],'1/cm',z[i],dim,factor*H2kJ*z[i],dimension))
        print('')
        print('%s %12.8f %s %9.3f %s' % ('  *** total ZPE:   ',zpe,dim,factor*H2kJ*zpe,dimension))
        print('  ----------------------------------------------------------')

###  end of printout24 ###

def printout26(name,dim,freqs,trans,rot,vib,f,BOTTOM,zpe,z):

    if( name == 'enthalpy' ):
        factor    = 1.0
        dimension = 'kJ/mol'
        print('  @@@ Thermal corrections to {0} {1}'.format(name,'@@@'))
    elif( name == 'entropy ' ):
        factor    = 1000.0
        dimension = 'J/K.mol'
        print('  @@@ Thermal corrections to {0} {1}'.format(name,'@@@'))
    elif( name == 'Heat capacity' ):
        factor    = 1000.0
        dimension = 'J/K.mol'
        print('  @@@ {0} {1}'.format(name,'(constant volume) @@@'))

    print('{0} {1:12.8f} {2} {3:9.3f} {4}'.format(' -- Translational: ',trans,dim,factor*H2kJ*trans,dimension))
    print('{0} {1:12.8f} {2} {3:9.3f} {4}'.format(' -- Rotational:    ',rot  ,dim,factor*H2kJ*rot  ,dimension))
    print(              '{0}'.format(' -- Vibrational:   '))
    if BOTTOM:
        print('      from the bottom of the valley')
    else:
        print('      from the first frequency')
    print('  *** individual frequency contrib. ***')
    print('')


    for i in range(len(freqs)):
        print('{0:12.2f} {1} {2:14.8f} {3} {4:9.3f} {5}'.format(freqs[i],'1/cm',f[i],dim,factor*H2kJ*f[i],dimension))
    print('')
    print('{0} {1:12.8f} {2} {3:9.3f} {4}'.format('  *** total vib:   ',vib,dim,factor*H2kJ*vib,dimension))
    print('')

    print('{0} {1:12.8f} {2} {3:9.3f} {4}'.format('  @@@ Total:       ',trans+rot+vib,dim,factor*H2kJ*(trans+rot+vib),dimension))
    print('  ----------------------------------------------------------')
    if( name == 'enthalpy'):
        print(              '{0}'.format(' -- ZPE:           '))
        print('  *** individual frequency contrib. ***')
        print('')

        for i in range(len(freqs)):
            print('{0:12.2f} {1} {2:14.8f} {3} {4:9.3f} {5}'.format(freqs[i],'1/cm',z[i],dim,factor*H2kJ*z[i],dimension))
        print('')
        print('{0} {1:12.8f} {2} {3:9.3f} {4}'.format('  *** total ZPE:   ',zpe,dim,factor*H2kJ*zpe,dimension))
        print('  ----------------------------------------------------------')

###  end of printout26 ###

def print_Q(temp,press,MW,rTs,sigma,linear,freqs,bottom): # it prints out the value of the partition functions
    if (P_ver == 24):
        print('Partition function for')
        print('%s %30.6f' %('-- translation',q_trans(temp,press,MW)))
        print('%s %30.6f' %('-- rotation   ',q_rot(rTs,temp,sigma,linear)))
    else:
        print('Partition function for')
        print('{0} {1:30.6f}'.format('-- translation',q_trans(temp,press,MW)))
        print('{0} {1:30.6f}'.format('-- rotation   ',q_rot(rTs,temp,sigma,linear)))
    
    Q_Vib = 1.0
    for i in range(len(freqs)):
        Q_Vib = Q_Vib*(q_vib(vib_Temp(freqs[i]),temp,bottom))

    if (P_ver == 24):
        print('%s %30.6f' % ('-- vibration  ',Q_Vib))
    else:
        print('{0} {1:30.6f}'.format('-- vibration  ',Q_Vib))
### end of print_Q ###
    
def print_E(): # just a warning
    print('')
    if (P_ver == 24):
        print('%s' % ('    ---   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!   ---'))
        print('%s' % ('          DO NOT forget the electronic contribution'))
        print('%s' % ('    ---   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!   ---'))
    else:
        print('{0}'.format('    ---   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!   ---'))
        print('{0}'.format('          DO NOT forget the electronic contribution'))
        print('{0}'.format('    ---   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!   ---'))
    print('')
### end of print_E ###

def spinorbit():
    print("@@@  C   @@@")
    q_elec(1.0,3.0,16.40,5.0,43.40,T298)          #C
    print("@@@  O   @@@")
    q_elec(5.0,3.0,158.265,1.0,226.977,T298)      #O
    print("@@@  F   @@@")
    q_elec(4.0,2.0,404.10,6.0,102405.71,T298)     #F
    print("@@@  Cl  @@@")
    q_elec(4.0,2.0,882.3515,6.0,71958.363,T298)   #Cl
    print("@@@  CF  @@@")
    q_elec(2.0,2.0,78.13298252,4.0,25000.0,T298)  #CF
    print("@@@  CCl @@@")
    q_elec(2.0,2.0,134.92,4.0,20000.0,T298)       #CCl
    print("@@@  ClO @@@")
    q_elec(2.0,2.0,320.31,2.0,31650,T298)         #ClO
    print("@@@  O2  @@@")
    q_elec(3.0,2.0,7882.39,1.0,13120.9085,T298)   #O2
    print("@@@  NO  @@@")
    q_elec(2.0,2.0,121.1,1.0,1000000.00,T298)     #NO
    print("@@@  S  @@@")
    q_elec(5.0,3.0,396.09,1.0,573.65,T298)        #S
    print("@@@  SH  @@@")
    q_elec(2.0,2.0,377.01,2.0,30663,T298)         #SH
    print("@@@  SO  @@@")
    q_elec(3.0,2.0,6350,1.0,10510,T298)           #SO


if __name__ == "__main__":
    main(sys.argv[1:])
